/* This class represents a transport connection between the gateway and the agent from the gateway's perspective */
import java.net.*;
import java.io.*;

import java.util.*;
import java.lang.*;

public class MGCPGateway {
  private int port;

  private String packetString;
  private StringTokenizer tokens;
  private long   randomId;

  private int callagentPort, packetLength;
  private InetAddress caAddress;

  ////////  MGCP PDU data //////////////////
  private ArrayList CommandHeader;
  /* This is for using in future */
  private ArrayList ResponseHeader;
  /* List waiting acknowledge of Retransmission object */
  private Vector yetAckList;
  /* all transaction Ids not acknowledged until now */
  private Vector transactionIdList;
  private String CallId;
  private String ConnectionId;
  private String retConnectionId;
  private String Mode;
  private String EndpointId;
  private String RequestIdentifier;
  private String RequestEvents;

  /* Local and Remote SDP information */
  private Sdp localSDPInformation, remoteSDPInformation;
  private connectionOptions localConnectionOptions;

  /* JMF Agent */
  private RTPAgent agent;

  private String Command, responseCode, transactionId;

  /* Pakcet Header Information */
  private String responseHeader, responseSDP;
  private String responseComment;

  private String returnString;
  private String AckHeader;

  /* Check whether Ack is Received
   *               Transaction Id is duplicated
   *               Connection Id is the same as previous ones
   */

  private boolean connectionEstablished;
  private boolean wrongCommand;
  private boolean received;
  private boolean duplicateId;
  private boolean sameConnId;
  private boolean presentSdp;

  /* this parameter is set up
  /* if call agent send connectionOptions parameter(L: ...)
   */
  private boolean receivedParameters;

  /* this parameter is set up by jmf */
  /* Now, jmf has no their api, it is set
   * if Call Agent send connectionOptions,
   * for testing parsing and working
   */
  private boolean sendingParameters;

  public MGCPGateway (int port) {
    this.port = port;
    Random random = new Random();
    yetAckList = new Vector();
    transactionIdList = new Vector();
    randomId = (long)(random.nextLong()%1000000000);
    if(randomId < 0)
      randomId = (-1) * randomId;
    connectionEstablished = false;
  }

  /* this is main method of MGCPGateway */
  public void execute () throws IOException {
    agent = new RTPAgent();
    DatagramSocket socket = new DatagramSocket (port);
    while (true) {
      DatagramPacket packet = receive (socket);
      if(packet.getLength() == 0) {
      	System.out.println("Empty Packet is received");
        continue;
      }

      /* Whenever it receives datagram from Call Agent
       * it initializes all variables
       */
      initializeStruct(packet);

      /* parse header
       * store header values to CommandHeader and some necessary variables
       * such as wrongCommand, Command(this is the command: CRCX...) etc.
       */
      parseHeader(new String(packet.getData()));

      /* If this Pakcet is made wrong, it is just ignored */
      if(wrongCommand == true) {
      	System.out.println("Wrong Datagram received");
        continue;
      }

      /* if received command is CRCX and a connection isn't already established
       * it recognize this datagram deserves to make a connection
       * but it is differentiated from K: id packet
       */
      if(!Command.equals("CRCX") && connectionEstablished == false)
        if(AckHeader.equals("")) {
          System.out.println("No Connection existed");
        }

      /* in case that received transactionId is the same as the previous one
       * which is not acknowledged yet.
       * it just ignore this datagarm because
       * retransmission function is sending the response datagram until geting ack.
       */
      if(duplicateId == true)
        continue;

      /* In case that the Received packet is "K: transactionId"
       * This was already processed in parseHeader function
       */
      if(AckHeader.equals("K:"))
        continue;
      /* after finishing parsing header of mgcp datagram
       * it parses the parameters and sdp
       */
      if(wrongCommand == false) {
        parseParameterSDP();
        if(sameConnId == true)
          callProcessRoutines();
      }

      /* it composes the response datagram
       * as following the received command and paramters
       */
      String responseToString = composeResponse();

      /* it re-transmits response datagram util being received the ack
       * for demo, transmission time is 3000
       * Originally it must be 30000
       * it uses resend thread whenever resend thread wakes up every 3 seconds,
       * resend thread checks acknowledgement list(yetAckList)
       * if ack is not received yet, resend thread retransmits the packet waiting ack.
       */
      Retransmission resend = new Retransmission(3000, this, responseToString);
      Acknowledge yetAck = new Acknowledge(transactionId, false, resend);
      yetAckList.addElement(yetAck);

      /* for checking transaction Id whether it's ack is received,
       *                                     the same transaction Id is received,
       */
      checkTransaction checkid = new checkTransaction(transactionId, responseToString, false);
      transactionIdList.addElement(checkid);

      responseToCA(responseToString);

      System.out.println(resend.toString());

      resend.start();

      if(sameConnId == true) {
        System.out.println("Agent Information in " + Command);
        System.out.println(agent.toString());
      }
      wrongCommand = false;
    }
  }

  /* to receive datagram from Call Agent */
  protected DatagramPacket receive (DatagramSocket socket) throws IOException {
    byte buffer[] = new byte[1024];
    DatagramPacket packet = new DatagramPacket (buffer, buffer.length);
    socket.receive (packet);
    return packet;
  }

  /* All data structure is initialized whenever every datagram is received */
  protected void initializeStruct(DatagramPacket Packet) {
    caAddress = Packet.getAddress();
    callagentPort = Packet.getPort();
    packetLength = Packet.getLength();
    CommandHeader = new ArrayList();
    ResponseHeader = new ArrayList();

    CallId = "";
    Mode = "";
    Command = "";
    responseCode = "";
    transactionId = "";
    AckHeader = "";
    retConnectionId = "";
    duplicateId = false;
    sameConnId = true;

    responseHeader = "";
    responseSDP = "";
    responseComment = "";
    returnString = "";

    RequestIdentifier = "";
    RequestEvents = "";
    wrongCommand = false;
    presentSdp = false;
    receivedParameters = false;
    sendingParameters = false;
  }

  /* MGCP header information is parsed
   * by following the command from MGCP Call Agent
   */
  protected void parseHeader(String packetString) throws IOException{
    tokens = new StringTokenizer(packetString.trim(), "\n\f\r");
    if(tokens.hasMoreTokens() != false) {
      String headerToken = tokens.nextToken();
      StringTokenizer headerTokenizer = new StringTokenizer(headerToken, " ,");
      String tempToken = headerTokenizer.nextToken();
      if(tempToken.equals("CRCX")){
      	Command = new String(tempToken);
      	CommandHeader.add(tempToken);
      	for(int count = 0; (count < 4) && headerTokenizer.hasMoreTokens(); count++) {
      	  tempToken = headerTokenizer.nextToken();
      	  if(count == 0) {
      	    transactionId = new String(tempToken);
      	    for(int i = 0; i < transactionIdList.size(); i++) {
      	      checkTransaction id = (checkTransaction)transactionIdList.elementAt(i);
      	      System.out.println(id.toString() + "\n");
      	      System.out.println("TransactionID: " + transactionId);
      	      if (transactionId.equals(id.getTransactionId())) {
      	      	responseToCA(id.getReturnString());
      	      	duplicateId = true;
      	      	System.out.println("Found the same Transaction Id");
      	      	break;
      	      }
      	    }
      	    if(duplicateId == false)
      	      CommandHeader.add(tempToken);
      	  }
        }
      }
      else if (tempToken.equals("MDCX")) {
      	Command = new String(tempToken);
      	CommandHeader.add(tempToken);
      	for(int count = 0; (count < 4) && headerTokenizer.hasMoreTokens(); count++) {
      	  tempToken = headerTokenizer.nextToken();
      	  if(count == 0) {
      	    transactionId = new String(tempToken);
      	    for(int i = 0; i < transactionIdList.size(); i++) {
      	      checkTransaction id = (checkTransaction)transactionIdList.elementAt(i);
      	      System.out.println(id.toString() + "\n");
      	      System.out.println("TransactionID: " + transactionId);
      	      if (transactionId.equals(id.getTransactionId())) {
      	      	responseToCA(id.getReturnString());
      	      	duplicateId = true;
      	      	System.out.println("Found the same Transaction Id");
      	      	break;
      	      }
      	    }
      	    if(duplicateId == false)
      	      CommandHeader.add(tempToken);
      	  }
        }
       	if(connectionEstablished == false)
      	  wrongCommand = true;
      }
      else if (tempToken.equals("DLCX")) {
      	Command = new String(tempToken);
      	CommandHeader.add(tempToken);
      	for(int count = 0; (count < 4) && headerTokenizer.hasMoreTokens(); count++) {
      	  tempToken = headerTokenizer.nextToken();
      	  if(count == 0) {
      	    transactionId = new String(tempToken);
      	    for(int i = 0; i < transactionIdList.size(); i++) {
      	      checkTransaction id = (checkTransaction)transactionIdList.elementAt(i);
      	      System.out.println(id.toString() + "\n");
      	      System.out.println("TransactionID: " + transactionId);
      	      if (transactionId.equals(id.getTransactionId())) {
      	      	responseToCA(id.getReturnString());
      	      	duplicateId = true;
      	      	System.out.println("Found the same Transaction Id");
      	      	break;
      	      }
      	    }
      	    if(duplicateId == false)
      	      CommandHeader.add(tempToken);
      	  }
        }
       	if(connectionEstablished == false)
      	  wrongCommand = true;

      }
      else if (tempToken.equals("K:")) {
      	AckHeader = tempToken;
        processAck(headerTokenizer);
      }
      else {
      	// for additional other commands
      }
    }
  }

  /* This function processes acknowledgement from Call Agent.
   * It always checks the transaction Id which is the same ID it waiting for.
   */
  protected void processAck(StringTokenizer headerTokenizer) {
    while(headerTokenizer.hasMoreTokens()) {
      String ackId = headerTokenizer.nextToken().trim();
      for(int i = 0; i < yetAckList.size(); i++) {
      	Acknowledge tempAck = (Acknowledge)yetAckList.elementAt(i);
      	if(ackId.equals(tempAck.getwaitAckID())) {
      	  tempAck.setReceived(true);
      	  System.out.println(tempAck.toString());

     	  tempAck.getRetThread().stopRequest();
      	  yetAckList.removeElementAt(i);
      	  break;
      	}
      }
      for(int j = 0; j < transactionIdList.size(); j++) {
        checkTransaction id = (checkTransaction)transactionIdList.elementAt(j);
      	if (ackId.equals(id.getTransactionId())) {
      	  transactionIdList.removeElementAt(j);
      	  break;
      	}
      }
    }
  }
  /* Parameter is parsed in this function */
  protected void parseParameterSDP() {
    while(tokens.hasMoreTokens()) {
      String ParameterSDPToken = tokens.nextToken("=:\r\f\n");
      String trimedString = new String(ParameterSDPToken.trim());

      /* parameter parsing */
      /* Call Identification Parameter */
      if(trimedString.equals("C"))
        CallId = new String(tokens.nextToken().trim());

        /* Parameter for receiving and sending mode */
      else if(trimedString.equals("M"))
        Mode = new String(tokens.nextToken().trim());

      /* In case of CreateConnection, this parameter should be abssent
       * In case of ModifyConnection and DeleteConnection,
       *            this parameter must be the same parameter as the returned value
       *            from MGCPGateway in CreateConnection.
       */
      else if(trimedString.equals("I")) {
        if(!Command.equals("CRCX")) {
          retConnectionId = new String(tokens.nextToken().trim());
          if(!ConnectionId.equals(retConnectionId.trim())) {
            sameConnId = false;
            System.out.println("Different ConnId");
          }
        } else {
           /* Wrong Paramter, This packet must be made wrongly */
           wrongCommand = true;
        }
      /* for piggybacking acknowledgment */
      } else if(trimedString.equals("K")) {
      	StringTokenizer acklist = new StringTokenizer(tokens.nextToken().trim(), " ");
      	processAck(acklist);
      /* Request Identifier Parameter */
      } else if(trimedString.equals("X"))  {
      	RequestIdentifier = new String(tokens.nextToken().trim());
      }
      /* Request Event Parameter */
      else if(trimedString.equals("R")) {
        RequestEvents = new String(tokens.nextToken().trim());
      }
      /* LocalConnectionOptions are included  */
      else if(trimedString.equals("L")) {
      	parseParameterL();
      	System.out.println(localConnectionOptions.toString());
      }
      /* in case of including sdp part */
      else if((presentSdp == true) && (trimedString.equals("v") == true))
        if(sameConnId == true)
          parseSDP();
    }
  }
  
  /* Parsing paramters used by the Call Agent to direct the handling of the connection by the gateway */
  protected void parseParameterL() {
    localConnectionOptions = new connectionOptions();

    String parameter = tokens.nextToken("\r\n");
    StringTokenizer parameterTokens = new StringTokenizer(parameter.trim(), ",");
    while (parameterTokens.hasMoreTokens()) {
      String optionsToken = parameterTokens.nextToken();
      StringTokenizer options = new StringTokenizer(optionsToken.trim(), ":");
      while(options.hasMoreTokens()) {
      	String whichParameter = options.nextToken().trim();

      	/* packetization period */
      	if(whichParameter.equals("p")) {
      	  String packetizationPeriod = options.nextToken();
      	  localConnectionOptions.setpacketPeriod(packetizationPeriod);
      	}
      	/* encoding method */
      	else if (whichParameter.equals("a")) {
      	  String encoding = options.nextToken();
      	  localConnectionOptions.setencodingMethod(encoding);
      	}
      	/* bandWidth */
      	else if (whichParameter.equals("b")) {
      	  String band = options.nextToken();
      	  localConnectionOptions.setbandWidth(band);
      	}
      	/* echo cancellation: on or off */
      	else if (whichParameter.equals("e")) {
      	  String echo = options.nextToken();
      	  localConnectionOptions.setechoCancellation(echo);
      	}
      	/* type of service */
      	else if (whichParameter.equals("t")) {
      	  String tty = options.nextToken();
      	  localConnectionOptions.settypeofService(tty);
      	}
      	/* silence suppresion: on or off */
      	else if (whichParameter.equals("s")) {
      	  String silence = options.nextToken();
      	  localConnectionOptions.setsilenceSuppression(silence);
      	}
      	/* gain control: auto or decimal number */
      	else if (whichParameter.equals("gc")) {
      	  String gc = options.nextToken();
      	  localConnectionOptions.setgainControl(gc);
      	}
      	/* reservation: g: guaranteed service, cl: controlled load, be: best effort */
      	else if (whichParameter.equals("r")) {
      	  String reserv = options.nextToken();
      	  localConnectionOptions.setreservationService(reserv);
      	}
      	/* encription key */
      	else if (whichParameter.equals("k")) {
      	  String ekey = options.nextToken();
      	  localConnectionOptions.setrtpSecurity(ekey);
      	}
      	/* type of network: IN, ATM or LOCAL */
      	else if (whichParameter.equals("nt")) {
      	  String ttn = options.nextToken();
      	  localConnectionOptions.settypeofNetwork(ttn);
      	}
      	else {
      	  //for other paramters if exists.
      	}
      }
    }
    receivedParameters = true;
  }
  /* If SDP infomation exists within datagram, this is the processing method */
  protected void parseSDP() {
    remoteSDPInformation = new Sdp();

    while (tokens.hasMoreTokens()) {
      String sdpToken = tokens.nextToken("=\n");
      System.out.println("Token " + sdpToken);
      if(sdpToken.equals("c")) {
      	StringTokenizer sdpTokenizer = new StringTokenizer(tokens.nextToken(), " ");
      	for(int count = 0; sdpTokenizer.hasMoreTokens(); count++) {
      	  if(sdpTokenizer.nextToken().equals("IP4"))
      	    remoteSDPInformation.setIPAddress(sdpTokenizer.nextToken());
      	}
      } else if (sdpToken.equals("m")) {
      	StringTokenizer sdpTokenizer = new StringTokenizer(tokens.nextToken(), " ");
      	for(int count = 0; sdpTokenizer.hasMoreTokens(); count++) {
      	  if(sdpTokenizer.nextToken().equals("audio"))
      	    remoteSDPInformation.setportNumber(Integer.parseInt(sdpTokenizer.nextToken()));
      	  // for video
      	  //else if(sdpTokenizer.nextToken().equals("video"))
      	    //System.out.println("Video");
      	    // for video communication
      	}
      } else if (sdpToken.equals("o")) {
      	StringTokenizer sdpTokenizer = new StringTokenizer(tokens.nextToken(), " ");
      	// For owner/creator and session identifier
      } else if(sdpToken.equals("s")) {
      	StringTokenizer sdpTokenizer = new StringTokenizer(tokens.nextToken(), " ");
      	// For session name
      } else if(sdpToken.equals("i")) {
      	// For session information
      } else if(sdpToken.equals("u")) {
      	// For uri of description
      } else if(sdpToken.equals("e")) {
      	// For e-mail address
      } else if(sdpToken.equals("p")) {
      	// For phone number
      } else if(sdpToken.equals("c")) {
      	// For connection information
      } else if(sdpToken.equals("b")) {
      	// For bandwidth information
      } else if(sdpToken.equals("z")) {
      	// For time zone adjustments
      } else if(sdpToken.equals("k")) {
      	// For enscription key
      } else if(sdpToken.equals("a")) {
      	// For session attribute lines
      } else
        continue;
    }
  }

  /* it call RTP interface function by following the command it received */
  protected void callProcessRoutines() {
    if (Command.equals("CRCX")) {
      ConnectionId = "" + randomId++;
      localSDPInformation = new Sdp();
      if (remoteSDPInformation == null) {
      	localSDPInformation.setIPAddress(agent.getLocalSessionAddress());
      	localSDPInformation.setportNumber(agent.getLocalSessionPort());
      } else {
      	agent.setRemoteSession(remoteSDPInformation.getIPAddress(), Integer.parseInt(remoteSDPInformation.getportNumber()));

      	localSDPInformation.setIPAddress(agent.getLocalSessionAddress());
      	localSDPInformation.setportNumber(agent.getLocalSessionPort());
      }
      if(receivedParameters == true) {
      	// call jmf routines
      	// receive sendingParameter values from jmf and
      	// set sendingParameters parameter to true if there are
      	sendingParameters = true;
      }
    } else if (Command.equals("MDCX")) {
      agent.setRemoteSession(remoteSDPInformation.getIPAddress(), Integer.parseInt(remoteSDPInformation.getportNumber()));
      if(receivedParameters == true) {
      	//call jmf routines
      	// receive sendingParameter values from jmf and
      	// set sendingParameters parameter to true if there are
      	sendingParameters = true;
      }
    } else if (Command.equals("DLCX")) {
    	agent.stopTransmission();
    } else {
    	// For other commands
    }
  }

  /* Compose the response to send to Call Agent */
  protected String composeResponse() {
    composeHeader();
    if(sameConnId == true)
      composeSDP();
    returnString = new String();
    returnString = responseHeader + responseSDP;
    return returnString;
  }

  /* Compose response header  */
  protected void composeHeader() {
    responseHeader = new String();
    responseComment = new String();
    if (Command.equals("CRCX")) {
      if(localSDPInformation.getIPAddress().equals("") &&
        (Integer.parseInt(localSDPInformation.getportNumber()) <= 0)) {
        responseHeader += 400;
        responseComment = " NOTOK" + "\n";
      }
      else {
        responseHeader += 200;
        responseComment = " OK" + "\n";
        connectionEstablished = true;
      }
    } else if (Command.equals("MDCX")) {
      if(wrongCommand == true) {
        responseHeader += 400;
        responseComment = " NOTOK" + "\n";
        return;
      } else if(!ConnectionId.equals(retConnectionId)) {
      	/* the received connection Id is different from the existing connection */
      	responseHeader += 515;
      	responseComment = " NOTOK" + "\n";
      } else {
        if(localSDPInformation.getIPAddress().equals("") &&
            (Integer.parseInt(localSDPInformation.getportNumber()) <= 0)) {
          responseHeader += 400;
          responseComment = " NOTOK" + "\n";
        }
        else {
          responseHeader += 200;
          responseComment = " OK" + "\n";
        }
      }
    } else if (Command.equals("DLCX")) {
      if(wrongCommand == true) {
        responseHeader += 400;
        responseComment = " NOTOK" + "\n";
      } else if(!ConnectionId.equals(retConnectionId)) {
      	/* the received connection Id is different from the existing connection */
      	responseHeader += 515;
      	responseComment = " NOTOK" + "\n";
      } else {
        responseHeader += 200;
        responseComment = " OK" + "\n";
        connectionEstablished = false;
      }
    }

    responseHeader += " ";
    responseHeader += (String)CommandHeader.get(1);
    responseHeader += responseComment;
    if(wrongCommand == true)
      return;
    if(sameConnId == true)
      responseHeader += "I:" + ConnectionId + "\r\n";
    else
      responseHeader += "I:" + retConnectionId + "\r\n";
    if(sendingParameters == true)
      responseHeader += "L: " + localConnectionOptions.toString();
  }

  /* Compose the SDP information using RTP Agent information */
  protected void composeSDP() {
    if(wrongCommand == true)
      return;
    responseSDP = new String();
    responseSDP += "v=0" + "\n";
    responseSDP += "c=IN IP4 " + localSDPInformation.getIPAddress() + "\n";
    responseSDP += "m=audio " + localSDPInformation.getportNumber() + "\n";
  }

  /* This is for debuging */
  protected void printHeader() {
    for (Iterator i = CommandHeader.iterator(); i.hasNext();)
      System.out.println(i.next());
  }

  /* This is for debuging */
  protected void printAll() {
    for (Iterator i = CommandHeader.iterator(); i.hasNext();)
      System.out.print(i.next() + " ");
    System.out.println("\nCallId: " + CallId);
    if(!ConnectionId.equals(""))
      System.out.println("ConnectionId: " + ConnectionId);
    if(!Mode.equals(""))
      System.out.println("Mode: " + Mode);
    if(localSDPInformation != null)
      System.out.println("SDP: " + localSDPInformation);
  }

  /* This is for debuging */
  public Vector getyetAckList() {
    return yetAckList;
  }

  /* Send the response to Call Agent */
  public void responseToCA (String ret) throws IOException {
    byte[] sendByte = ret.getBytes();
    DatagramPacket response = new DatagramPacket
      (sendByte, sendByte.length, caAddress, callagentPort);
    DatagramSocket socket = new DatagramSocket ();

    socket.send (response);
    socket.close();
  }

  /* To check whether the ack is received
   * If not, it retransmit the response waiting for ack
   */
  public synchronized void retransmissionCall (String retString) {
    try {
      int i;
      for(i = 0; i < yetAckList.size(); i++) {
      	Acknowledge waitingAck = (Acknowledge)yetAckList.elementAt(i);
      	if(waitingAck.getReceived()) {
      	  waitingAck.getRetThread().stopRequest();
      	  yetAckList.removeElementAt(i);
      	  break;
      	}
      	else {
      	  System.out.println ("Retransmiting!");
      	  responseToCA (retString);
      	  break;
      	}
      }
    } catch (IOException ex) {
      ex.printStackTrace ();
    }
  }

  public static void main (String[] args) throws IOException {
    if (args.length != 1)
      throw new IllegalArgumentException ("Syntax: MGCPGateway <port>");
    MGCPGateway server = new MGCPGateway (Integer.parseInt (args[0]));
    server.execute ();
  }

}